import TSMFull
reload(TSMFull)
from TSMFull import *

import utilities
reload(utilities)
from utilities import *

# import DrawingWindowStandalone
# reload(DrawingWindowStandalone)
# from DrawingWindowStandalone import *

class Map:
    # Takes list of boxes making up the obstacles in the world
    def __init__(self, obstacleBoxes):
        self.boxes = obstacleBoxes
        self.segs = []
        for (x1,y1,x2,y2) in obstacleBoxes:
            self.segs.extend((((x1,y1),(x1,y2)),
                              ((x1,y2),(x2,y2)),
                              ((x2,y2),(x2,y1)),
                              ((x2,y1),(x1,y1))))

    def draw(self, window):
        for (x1,y1,x2,y2) in self.boxes:
            window.drawRect((x1,y1),(x2,y2), color = "gray")

class D2GridWorld(PrimitiveSM):
    pass

class D2GridWorldFromMap(D2GridWorld):
    robotWidth = 0.2
    
    def __init__(self, theMap, xmin, xmax, ymin, ymax, gridN):
        self.xmin = float(xmin)
        self.xmax = float(xmax)
        self.ymin = float(ymin)
        self.ymax = float(ymax)
        self.gridN = gridN
        self.xstep = (self.xmax - self.xmin)/self.gridN
        self.ystep = (self.ymax - self.ymin)/self.gridN
        self.map = theMap

        # Is the grid cell centered at world coordinates (x,y) inside
        # the box?  It's not enough to see if the point (x,y) is in
        # the box: we have to check whether or not any part of the
        # whole cell is in it.  Furthermore, we have to increase the
        # cell size by the radius of the robot (we're assuming it's
        # round).
        def contains((x1,y1,x2,y2),x,y,margin):
            xmin = min(x1,x2)
            ymin = min(y1,y2)
            xmax = max(x1,x2)
            ymax = max(y1,y2)
            return x >= xmin - self.xstep/2.0 - margin and \
                   x <= xmax + self.xstep/2.0 + margin and \
                   y >= ymin - self.ystep/2.0 - margin and \
                   y <= ymax + self.ystep/2.0 + margin

        # Is the grid cell whose center is at real-world coordinates
        # (x,y) occupied in this map?
        def occupied(x,y,obstacleBoxes):
            for obstacle in obstacleBoxes:
                if contains(obstacle,x,y,self.robotWidth):
                    return True
            return False

        self.freeSpace = make2DArray(self.gridN, self.gridN, 1)
        for xIndex in range(self.gridN):
            for yIndex in range(self.gridN):
                (px,py) = self.indicesToPose((xIndex,yIndex))
                if occupied(px,py,theMap.boxes):
                    self.freeSpace[xIndex][yIndex] = 0

        # Just allow compass motions and nop in every state
        def possibleActions(s):
            return [(0, 0), (-1, 0), (+1, 0), (0, +1), (0, -1)]

        # If you try to move into a wall, you stay where you are.
        def transitionFn((ix, iy), (dx, dy)):
            newix = ix + dx
            newiy = iy + dy
            if self.free((newix, newiy)):
                return (newix, newiy)
            else:
                return (ix, iy)

        def outputFn(s):
            return s

        # default initial state is (1,1), but that might not even be free. 
        PrimitiveSM.__init__(self, transitionFn, outputFn,
                             lambda x: (1,1), legalInputs = possibleActions)

    def free(self, (i,j)):
        return i >= 0 and i < self.gridN and j >= 0 and j < self.gridN \
               and self.freeSpace[i][j] == 1
    
    def xToIndex(self, x):
        return min(int(round(((x-self.xmin)/(self.xmax-self.xmin)) \
                             *self.gridN)),self.gridN-1)
    def yToIndex(self, y):
        return min(int(round(((y-self.ymin)/(self.ymax-self.ymin))  \
                             *self.gridN)),self.gridN-1)

    def poseToIndices(self, (x,y)):
        return(self.xToIndex(x),self.yToIndex(y))

    def indicesToPose(self, (ix,iy)):
        return(self.xmin+(float(ix)*(self.xmax-self.xmin)/self.gridN),
               self.ymin+(float(iy)*(self.ymax-self.ymin)/self.gridN))

    def drawWorld(self, window, windowWidth):
        boxDim = int(self.xstep * (windowWidth / (self.xmax - self.xmin + 1)))\
                 - 4
        window.clear()
        x = self.xmin
        for xIndex in range(self.gridN):
            y = self.ymin
            for yIndex in range(self.gridN):
                if not self.free((xIndex,yIndex)):
                    window.drawRobot(x, y, x, y, color = "black",
                                     size = boxDim)
                else:
                    window.drawRobot(x, y, x, y, color = "white",
                                     size = boxDim)
                y = y + self.ystep
            x = x + self.xstep
        # Draw the map boxes
        self.map.draw(window)

    def drawState(self, indices, window, windowWidth, color, label = None):
        boxDim = int(self.xstep * (windowWidth / (self.xmax - self.xmin + 1)))\
                 - 4
        (x, y) = self.indicesToPose(indices)
        window.drawRobot(x, y, x, y, color = color, size = boxDim)
        if label:
            window.drawText(x, y, label)


######################################################################
###
###    Some worlds
###
######################################################################
    

# Ranges: 0.0, 3.05, 0.0, 3.05
shannonBoxes = ((0,0,0,3.05),(0,3.05,2.03,3.05),(2.03,0,3.05,3.05),\
                (0,0,2.03,0), (.51,2.24,1.12,1.88))

# Ranges: 0.0, 4.0, 0.0, 4.0
sheBoxes = ((0,0,0,4),(0,4,4,4),(4,0,4,4),(0,0,4,0),(1,0,1,2),(1,2,2,2),
            (2,1,3,1),(3,1,3,3),(3,3,1,3))

# Ranges: 0.0, 4.0, 0.0, 4.0
openBoxes = ((0,0,0,4),(0,4,4,4),(4,0,4,4),(0,0,4,0))

# Ranges: 0.0, 4.0, 0.0, 4.0
tutorialBoxes = ((0,0,0,4),(0,4,4,4),(4,0,4,4),(0,0,4,0),\
                 (0,4,1,3),(2,0,2,2),(2,2,3,2))

# Ranges: 0.0, 3.05, 0.0, 3.05
playpenBoxes = ((0,0,0,3.05),(0,3.05,2.03,3.05),(2.03,0,3.05,3.05),\
                (0,0,2.03,0))

# Ranges: 0.0, 4.0, 0.0, 3.0
kjBoxes = ((0, 0, 0, 3), (0, 3, 4, 3), (4, 3, 4, 0), (4, 0, 0, 0), \
           (1, 0, 1, 1), (1, 1, 2, 1), (3, 1, 3, 2), \
           (1, 2, 3, 2))

braveBoxes = ((0, 0, 0, 7), (0, 7, 7, 7), (7, 7, 7, 0), (7, 0, 0, 0), \
              (2.,0.,7.,0.), (2.,0.,2.,1.), (1.,1.,2.,1.), (1.,1.,1.,2.),\
              (0.,2.,1.,2.), (0.,3.,2.,3.), (2.,2.,2.,4.), (2.,2.,3.,2.),\
              (3.,2.,3.,1.), (3.,1.,4.,1.), (4.,1.,4.,2.), (0.,5.,3.,5.),\
              (3.,3.,3.,5.), (3.,3.,5.,3.), (5.,1.,5.,3.), (5.,1.,7.,1.))


