from soar.widgets import *

import GridWorld
reload(GridWorld)
from GridWorld import *

import Planner
reload(Planner)
from Planner import *

import XYDriver
reload(XYDriver)
from XYDriver import *

import TSMFull
reload(TSMFull)
from TSMFull import *

# For knowing our exact location, in the simulator only
def cheatPose():
    return app.soar.output.abspose.get()    

# Plan is a list of (action, state) pairs
def drawPlan(gw, plan):
    gw.drawWorld(robot.w, robot.windowWidth)
    lastStep = len(plan) - 1
    for step, (a, s) in enumerate(plan):
        if step == 0:
            color = "DarkGreen"
        elif step == lastStep:
            color = "Gold"
        else:
            color = "Blue"
        gw.drawState(s, robot.w, robot.windowWidth, color)

## This is a terminating state machine class.  It takes a goal, a
## grid-world, and a planner as argument at initialization time.  The
## state machine plans a path from the current state to the goal, and
## executes a terminating behavior to drive to the first "subgoal"
## location on that path.  When that subgoal is reached, or the driver
## terminates for another reason, the path is re-planned, and then the
## robot drives to the first subgoal on the new path.  When the goal
## location is reached, the behavior terminates.

## IPE stands for "interleaved planning and execution"        
class IPE():

    def __init__(self, goalIndices, gw, planner):
        self.goalIndices = goalIndices
        self.gw = gw
        self.planner = planner

    def start(self, input):
        (sonars, (x, y, theta)) = input
        # Get grid indices for robot's x,y location
        self.currentIndices = self.gw.poseToIndices((x,y))
        # Make a plan
        plan = self.planner.plan(initialState = self.currentIndices,
                                 goalTest = lambda x: x == self.goalIndices)
        if plan:
            drawPlan(self.gw, plan)
        else:
            # Couldn't get there:  draw start and goal
            drawPlan(self.gw, [(None, self.currentIndices),
                               (None, self.goalIndices)])
            
        if plan and len(plan) > 1:
            # Return a terminating behavior to take the first action
            # plan[1] is the first real step,  [0] is the action part
            motion = self.gw.indicesToPose(plan[1][0])
            self.macroAction = XYDriverOA(motion, robot.goalEpsilon)
        else:
            # Wait 5 steps and see if we've been kidnapped
            self.macroAction = Wait(5)

        # Start the macroAction
        self.macroAction.start(input)

    def step(self, input):
        (sonars, (x, y, theta)) = input
        if self.macroAction.done():
            self.start(input)
        self.currentIndices = self.gw.poseToIndices((x,y))
        return self.macroAction.step(input)

    def currentOutput(self):
        return self.macroAction.currentOutput()

    def done(self):
        return self.currentIndices == self.goalIndices

######################################################################
###
###          Brain methods
###
######################################################################
def setup():
    (xmin, xmax, ymin, ymax) = (0.0, 4.0, 0.0, 4.0)
    (windowWidth, windowHeight) = (400, 400)
    robot.windowWidth = windowWidth # we need this elsewhere
    gridN = 20
    robot.goalEpsilon = (xmax - xmin) / (5 * gridN)
    
    gw = D2GridWorldFromMap(Map(sheBoxes), xmin, xmax, ymin, ymax, gridN)
    gwPlanner = Planner(gw)
    robot.w = DrawingWindow(windowWidth, windowHeight,
                            xmin-0.5, xmax+0.5, ymin-0.5, ymax+0.5, "Plan")
    gw.drawWorld(robot.w, windowWidth)

    robot.driver = IPE(gw.poseToIndices((3.5, 0.5)), gw, gwPlanner)
    robot.driver.start(collectSensors())

def step():
    if robot.driver.done():
        print "done"
        doAction((0.0, 0.0))
    else: 
        doAction(robot.driver.step(collectSensors()))

def collectSensors():
    return [sonarDistances(), cheatPose()]
    
def doAction(a):
    speedLimit = 1.0
    def clip(v):
        return min(max(v, -speedLimit), speedLimit)
    (fvel, rvel) = a
    if fvel < 0.0:
        print "Going backwards!", fvel
    motorOutput(clip(fvel), clip(rvel))

def shutdown():
    if robot.w:
        robot.w.destroy()

