import utilities
reload(utilities)
from utilities import *

import TSMFull
reload (TSMFull)
from TSMFull import *

# Define a new terminating state machine that goes to an (x,y) pose
# relative to the pose when start is called
class XYDriver:
    forwardGain = 3.0 #1.5
    rotationGain = 5.0 #1.5
    angleEps = 0.05

    def __init__(self, deltaXY, distEps):
        self.deltaXY = deltaXY
        self.distEps = distEps

    def start(self, input):
        (sonars, (px, py, ptheta)) = input
        (dx, dy) = self.deltaXY
        self.goalXY = (px + dx, py + dy)
        self.currentPose = (px, py, ptheta)
    
    def step(self, (sonars, pose)):
        self.currentPose = pose
        return self.currentOutput()

    def currentOutput(self):
        (px, py, ptheta) = self.currentPose
        # The location we're trying to reach, described in coordinate
        # system of the robot's odometry
        # The direction from the robot's current location to the desired
        # location
        headingTheta = angleToPoint((px,py),self.goalXY)
        # The difference between the desired and actual headings
        headingDeltaTheta = fixAnglePlusMinusPi(headingTheta - ptheta)

        if nearAngle(ptheta, headingTheta, self.angleEps):
            # The heading is close to right, so we should drive forward
            # r is the distance to the goal
            r = poseDist2((px, py), self.goalXY)
            # Move at a speed proportional to the distance
            return (r * self.forwardGain, 0.0)
        else:
            # We need to adjust the heading before moving.  Rotate at a
            # speed proportional to the heading error
            return (0.0, headingDeltaTheta * self.rotationGain)

    # We don't care what our orientation is as long as our (x,y)
    # coordinates are close  
    def done(self):
        return nearPose2(self.goalXY, self.currentPose[:-1], self.distEps)

# Define a new terminating behavior that goes to an (x,y) pose
# Terminate if blocked.  Have to check for (and remember) whether
# we're blocked when we start, and on each step.

blockedThreshold = 0.1

class XYDriverOA (XYDriver):
    def start(self, (sonars, pose)):
        self.blocked = min(sonars) < blockedThreshold
        XYDriver.start(self, (sonars, pose))

    def step(self, (sonars, pose)):
        self.blocked = min(sonars) < blockedThreshold
        return XYDriver.step(self, (sonars, pose))

    # Terminate if we're too close to any obstacle
    def done(self):
        return XYDriver.done(self) or self.blocked

class Stop:
    def __init__(self):
        pass
    def start(self, input = None):
        pass
    def step(self, input = None):
        return self.currentOutput()
    def currentOutput(self):
        return (0.0, 0.0)
    def done(self):
        return True

class Wait:
    def __init__(self, waitingTime):
        self.waitingTime = waitingTime
    def start(self, input = None):
        self.count = self.waitingTime
    def step(self, input = None):
        print "waiting"
        self.count = self.count - 1
        return self.currentOutput
    def currentOutput(self):
        return (0.0, 0.0)
    def done(self):
        return self.count <= 0


# Drive forward a distance d from the pose at initialization time
class ForwardTSM:
    forwardGain = 3.0
    distTargetEpsilon = 0.05

    def __init__(self, delta):
        self.deltaDesired = delta
        
    def done(self):
        return within(poseDist2(self.startingPose, self.currentPose),
                      self.deltaDesired,
                      self.distTargetEpsilon)
    
    def step(self, (sonars, (x, y, theta))):
        self.currentPose = (x, y)
        return self.currentOutput()
    
    def start(self, (sonars, (x, y, theta))):
        print "Starting forward", self.deltaDesired
        self.startingPose = (x, y)
        self.currentPose = (x, y)
        
    def currentOutput(self):
        error = self.deltaDesired - \
                     poseDist2(self.startingPose, self.currentPose)
        return (self.forwardGain * error, 0)

class BackwardTSM (ForwardTSM):
    def currentOutput(self):
        error = self.deltaDesired - \
                     poseDist2(self.startingPose, self.currentPose)
        return (-self.forwardGain * error, 0)    

class RotateTSM:
    rotationalGain = 3.0
    angleEpsilon = 0.1
    
    def __init__(self, headingDelta):
        self.headingDelta = headingDelta
        
    def done(self):
        return nearAngle(self.desiredHeading, self.currentHeading,
                         self.angleEpsilon)
    
    def step(self, (sonars, (x, y, theta))):
        self.currentHeading = theta
        return self.currentOutput()
    
    def start(self, (sonars, (x, y, theta))):
        print "Starting to rotate", self.headingDelta
        self.currentHeading = theta
        self.desiredHeading = fixAnglePlusMinusPi(theta + self.headingDelta)

    def currentOutput(self):
        return (0, self.rotationalGain * \
                fixAnglePlusMinusPi(self.desiredHeading - self.currentHeading))

stepLength = 0.1

def dance():
    return RepeatTSM(SequentialTSM([ForwardTSM(stepLength),
                                    BackwardTSM(stepLength),
                                    RotateTSM(math.pi / 2)]), 4)
